// IFSDoc.cpp : implementation of the CIFSDoc class
//

#include "stdafx.h"
#include "IFS.h"

#include "IFSDoc.h"
#include "ProgressDlg.h"
#include "fstream.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CIFSDoc

IMPLEMENT_DYNCREATE(CIFSDoc, CDocument)

BEGIN_MESSAGE_MAP(CIFSDoc, CDocument)
	//{{AFX_MSG_MAP(CIFSDoc)
	ON_COMMAND(ID_PARAMS, OnParams)
	ON_COMMAND(ID_SAVEASBMP, OnSaveasbmp)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CIFSDoc construction/destruction

CIFSDoc::CIFSDoc()
{
	// TODO: add one-time construction code here
	bm = NULL;
}

CIFSDoc::~CIFSDoc()
{
}

BOOL CIFSDoc::OnNewDocument()
{
	if (!CDocument::OnNewDocument())
		return FALSE;

	// TODO: add reinitialization code here
	// (SDI documents will reuse this document)

	dlg.m_Auto = true;
	dlg.m_Iterations = 1;
	dlg.m_ImageSize = 1;
	for (int i = 0; i < 6; i++)
	{
		dlg.use[i] = 0;
		for (int j = 0; j < 6; j++)
			dlg.coef[i][j] = 0.0;
	}
	dlg.use[0] = 1;
	dlg.coef[0][0] = 1.0;
	dlg.coef[0][3] = 1.0;

	if (bm) delete bm;
	bm = NULL;

	return TRUE;
}



/////////////////////////////////////////////////////////////////////////////
// CIFSDoc serialization

void CIFSDoc::Serialize(CArchive& ar)
{
	if (ar.IsStoring())
	{
		// TODO: add storing code here
		ar.Write(dlg.use, sizeof(int) * 6);
		ar.Write(dlg.coef, sizeof(double) * 36);
		ar << dlg.m_ImageSize;
		ar << dlg.m_Auto;
		ar << dlg.m_Iterations;
	}
	else
	{
		// TODO: add loading code here
		ar.Read(dlg.use, sizeof(int) * 6);
		ar.Read(dlg.coef, sizeof(double) * 36);
		ar >> dlg.m_ImageSize;
		ar >> dlg.m_Auto;
		ar >> dlg.m_Iterations;
		Calculate();
	}
}

/////////////////////////////////////////////////////////////////////////////
// CIFSDoc diagnostics

#ifdef _DEBUG
void CIFSDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CIFSDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CIFSDoc commands

void CIFSDoc::OnParams() 
{
	// TODO: Add your command handler code here
	double coef[6][6];
	int use[6], i, j, i1, i2, i3;

	for (i = 0; i < 6; i++)
		memcpy(coef[i], dlg.coef[i], sizeof(double) * 6);
	memcpy(use, dlg.use, sizeof(int) * 6);
	i1 = dlg.m_ImageSize;
	i2 = dlg.m_Iterations;
	i3 = dlg.m_Auto;

	if (dlg.DoModal() != IDOK) return;	

	for (i = 0; i < 6; i++)
	{
		if (use[i] != dlg.use[i])
		{
			SetModifiedFlag();
			break;
		}
		for (j = 0; j < 6; j++)
			if (coef[i][j] != dlg.coef[i][j])
			{
				SetModifiedFlag();
				break;
			}
	}
	if (i1 != dlg.m_ImageSize ||
		i2 != dlg.m_Iterations ||
		i3 != dlg.m_Auto)
		SetModifiedFlag();

	Calculate();
}

void CIFSDoc::Calculate()
{
	unsigned char *p, *p2, *p3, color;
	int size2, size3, i, i2, i3, j, m, n, iter, x, y;
	double a, b, c, d, e, f;

	bits = dlg.m_ImageSize + 6;
	size = 1 << bits;
	size2 = size * size;
	size3 = size2 >> 3;

	if (dlg.m_Auto)
		iter = bits << 1;
	else
		iter = dlg.m_Iterations;

	p = new unsigned char[size2];
	p2 = new unsigned char[size2];
	p3 = new unsigned char[size3];

	memset(p, 0, size2);
	CProgressDlg pd;
	pd.Create(IDD_PROGRESSDLG);

	for (n = 0; n < iter; n++)
	{
		memset(p2, 1, size2);
		for (m = 0; m < 6; m++)
		{
			if (!dlg.use[m]) continue;
			a = dlg.coef[m][0];
			c = dlg.coef[m][2];
			e = dlg.coef[m][4] * size;
			f = dlg.coef[m][5] * size;
			for (i = 0; i < size; i++)
			{
				i2 = i << bits;
				b = dlg.coef[m][1] * i + e;
				d = dlg.coef[m][3] * i + f;
				for (j = 0; j < size; j++)
				{
					color = p[i2 + j];

					x = int(a * j + b);
					y = int(c * j + d);

					if (x < 0 || x >= size || y < 0 || y >= size) 
						continue;

					p2[(y << bits) + x] &= color;
				}
				i2 += size;
			}
		}
		memcpy(p, p2, size2);
		pd.m_Progress.SetPos(n * 100 / iter);
	}

	pd.DestroyWindow();

	for (i = 0, i3 = (size - 1) << bits; i < size3; i += size >> 3)
	{
		i2 = i3;
		for (j = 0; j < size >> 3; j++)
		{
			p3[i + j] = 
				(p[i2] << 7) |
				(p[i2 + 1] << 6) |
				(p[i2 + 2] << 5) |
				(p[i2 + 3] << 4) |
				(p[i2 + 4] << 3) |
				(p[i2 + 5] << 2) |
				(p[i2 + 6] << 1) |
				p[i2 + 7];
			i2 += 8;
		}
		i3 -= size;
	}

	if (bm) delete bm;
	bm = new CBitmap();
	bm->CreateBitmap(size, size, 1, 1, p3);

	delete p3;
	delete p2;
	delete p;

	UpdateAllViews(NULL);
}

void CIFSDoc::OnSaveasbmp() 
{
	// TODO: Add your command handler code here
	if (!bm) return;
	
	CFileDialog s(false, "bmp", NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
		" Windows Bitmap (*.bmp)|*.bmp|  (*.*)|*.*||", AfxGetMainWnd());

	if (s.DoModal() != IDOK) return;

	BITMAPFILEHEADER bfh;
	BITMAPINFOHEADER bih;
	RGBQUAD color0, color1;
	int img_size = size * size >> 3;
	int offs = sizeof(bfh) + sizeof(bih) + sizeof(RGBQUAD) * 2;
	int size3 = size >> 3;
	int size2 = img_size - size3;
	char *bits = new char[img_size];
	char *buf = new char[size3];
	bm->GetBitmapBits(img_size, bits);

	for (int i = size2; i > size2 >> 1; i -= size3)
	{
		memcpy(buf, bits + size2 - i, size3);
		memcpy(bits + size2 - i, bits + i, size3);
		memcpy(bits + i, buf, size3);
	}

	delete buf;	

	bfh.bfType = 0x4D42;
	bfh.bfSize = offs + img_size;
	bfh.bfReserved1 = 0;
	bfh.bfReserved2 = 0;
	bfh.bfOffBits = offs;

	bih.biSize = sizeof(bih);
	bih.biWidth = size;
	bih.biHeight = size;
	bih.biPlanes = 1;
	bih.biBitCount = 1;
	bih.biCompression = 0;
	bih.biSizeImage = img_size;
	bih.biXPelsPerMeter = 0;
	bih.biYPelsPerMeter = 0;
	bih.biClrUsed = 0;
	bih.biClrImportant = 0;

	memset(&color0, 0, sizeof(RGBQUAD));
	color1.rgbRed = 0xFF;
	color1.rgbGreen = 0xFF;
	color1.rgbBlue = 0xFF;
	color1.rgbReserved = 0;

	ofstream out(s.GetPathName(), ios::binary);
	out.write((char *)&bfh, sizeof(bfh));
	out.write((char *)&bih, sizeof(bih));
	out.write((char *)&color0, sizeof(RGBQUAD));
	out.write((char *)&color1, sizeof(RGBQUAD));
	out.write(bits, bih.biSizeImage);
	out.close();

	delete bits;
}
